
// http://arduino.stackexchange.com/questions/12587/how-can-i-handle-the-millis-rollover


void onPulseChange() {

  pulseData.dirtyPulseCount++;
  
  //overflows after ~50 days
  unsigned long currentTime = millis();
  unsigned long millisToCommit = pulseData.lastPulseCommittedAt - currentTime;
  int currentState = digitalRead(PULSE_PIN);



  if (millisToCommit <= min_pulsewidth && 
        pulseData.lastPulseCommittedAt != currentTime) {
    //Two interrupts can happen in the same millisecond! (<=)
    
    
    if (currentState != pulseData.lastPulseToCommitState) {
      //Serial.println("-");
      
       //Going back to original state which equals currentState, undo previously counted value
       pulseData.cleanPulseCount --;
       
       pulseData.lastPulseCommittedAt = pulseData.penultimatePulseCommittedAt;
       pulseData.lastPulseToCommitState = currentState;
       pulseData.dbg_confirmationsUndone++;
       
       //Clean pulse count is now confirmed as only one pulse count can be undone
      
    } else {
      // This is a confirmation of the state in previous interrupt, ignore
      //Serial.println("c");
      pulseData.dbg_confirmationsBeforeCommit++;
      
    }
    
    

    
  } else {
    //Interrupt after commit of previous pulse


    if (currentState != pulseData.lastPulseToCommitState) {
      //Serial.println("+");
      
      //New state after a long time, count
      pulseData.cleanPulseCount++;
      
      pulseData.penultimatePulseCommittedAt = pulseData.lastPulseCommittedAt;
      pulseData.lastPulseCommittedAt = currentTime + min_pulsewidth;
      pulseData.lastPulseToCommitState = currentState;
      
    } else {
      //Serial.print("m");
      //This means that we probably have not seen a state change, 
      //See it as a confirmation of previous long standing state, not counting
      pulseData.dbg_missedStateChanges ++;

    
    }
 
  }
  

}

void printPulseData() {
  Serial.print("dirtyPulseCount=");
  Serial.print(pulseData.dirtyPulseCount);
  
  Serial.print(", cleanPulseCount=");
  Serial.print(pulseData.cleanPulseCount);

  Serial.print(", lastPulseCommittedAt=");
  Serial.print(pulseData.lastPulseCommittedAt);

  Serial.print(", lastPulseToCommitState=");
  Serial.print(pulseData.lastPulseToCommitState);

  Serial.print(", penultimatePulseCommittedAt=");
  Serial.print(pulseData.penultimatePulseCommittedAt);

  Serial.print(", dbg_confirmationsUndone=");
  Serial.print(pulseData.dbg_confirmationsUndone);
  
  Serial.print(", dbg_confirmationsBeforeCommit=");
  Serial.print(pulseData.dbg_confirmationsBeforeCommit);

  Serial.print(", dbg_missedStateChanges=");
  Serial.print(pulseData.dbg_missedStateChanges);

  Serial.print(", lastSentCleanPulseCount=");
  Serial.print(pulseData.lastSentCleanPulseCount);

  Serial.print(", lastPulseCountSentAt=");
  Serial.print(pulseData.lastPulseCountSentAt);  
  Serial.println();
}


//Send emonPi data to Pi serial /dev/ttyAMA0 using struct JeeLabs RF12 packet structure
void send_as_rf12_demo_packet()   
{
  byte binarray[sizeof(rfm2pi)];
  memcpy(binarray, &rfm2pi, sizeof(rfm2pi));
  
  Serial.print("OK ");
  Serial.print(pulse_nodeid);
  for (byte i = 0; i < sizeof(binarray); i++) {
    Serial.print(' ');
    Serial.print(binarray[i]);
  }
  Serial.print(" (-0)");
  Serial.println();
  
  delay(10);

  pulseData.lastPulseCountSentAt = millis();
  
}
